/*

######################################################
###        simple LS algorithms for the SMWC       ###
######################################################

      Version: 1.0
      File:    ls.c
      Author:  Thomas Stuetzle
      Purpose: implementation of constructive and perturbative local search routines
      Check:   README and gpl.txt
      Remark:  the code is very basic and slow (no speed-ups are implemented)
      Copyright (C) 2004  Thomas Stuetzle
*/

#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <limits.h>

#include "global.h"
#include "ls.h"
#include "utilities.h"
#include "timer.h"

/* arrays keeping instance data */

long int   proc_time[MAX_JOBS];
long int   weight[MAX_JOBS];
long int   release_date[MAX_JOBS];

seq_struct sequence;

static long int  random_vector[MAX_JOBS];




void print_sequence(long int k) {
    long int i;
  
    printf("Objective function value = %ld\n", sequence.value);
    for( i = 0 ; i < number_jobs ; i++ ) {
	printf("%ld ", sequence.job_at_pos[i]); 
	if ( !(i % 20) )
	    printf("\n");
    }
}



void print_completion_times( ) {
    long int i;
  
    printf("Objective function value = %ld\n", sequence.value);
    for( i = 0 ; i < number_jobs ; i++ ) {
	printf("%ld ", sequence.completion_time_pos[i]); 
	if ( !(i % 20) )
	    printf("\n");
    }
}



void construct_sequence_random( ) {
    long int i;
  
    VERBOSE ( printf("Construct random solution\n"); )
    generate_random_vector();
    for( i = 0 ; i < number_jobs ; i++ ) {
	sequence.job_at_pos[i] = random_vector[i];
	sequence.pos_of_job[random_vector[i]] = i;
    }
    sequence.value = compute_obj_function( );
}



void construct_sequence_canonical( ) {
    long int i;
  
    VERBOSE ( printf("Build canonical sequence\n"); )
    for( i = 0 ; i < number_jobs ; i++ ) {
	sequence.job_at_pos[i] = i;
	sequence.pos_of_job[i] = i;
    }
    sequence.value = compute_obj_function( );
}



void construct_sequence_WSPT( ) {
    long int i, j;
    float    ratio_weight_proc[MAX_JOBS];
    long int numbers[MAX_JOBS];
    long int job_used[MAX_JOBS];
    long int start_time_job, job = 0, job_found, added_job = 0, help, index = 0;

    VERBOSE ( printf("Construct solution WSPT\n"); )
    for( i = 0 ; i < number_jobs ; i++ ) {
	ratio_weight_proc[i] = (float)weight[i] / (float)proc_time[i]; 
	numbers[i] = i;
	job_used[i] = FALSE;
    }

    sort2( ratio_weight_proc, numbers, 0, number_jobs - 1 );

    start_time_job = 0;
    for( i = 0 ; i < number_jobs ; i++ ) {
	job_found = FALSE;
	for( j = 0 ; j < number_jobs ; j++ ) {
	    job =  numbers[j];
	    if ( !(job_used[j]) && (release_date[job] <= start_time_job) ) {
		sequence.job_at_pos[i] = job;
		sequence.pos_of_job[job] = i;
		job_used[j] = TRUE;
		job_found = TRUE;
		added_job = job;
		break;
	    }
	}
	if ( !job_found ) {
	    help = LONG_MAX;
	    for( j = 0 ; j < number_jobs ; j++ ) {
		job =  numbers[j];
		if ( release_date[job] < help && !job_used[j] ) {
		    added_job = job;
		    index = j;
		    help = release_date[job];
		}
	    }
	    sequence.job_at_pos[i] = added_job;
	    sequence.pos_of_job[added_job] = i;
	    job_used[index] = TRUE;
	    start_time_job = release_date[added_job] + proc_time[added_job];
	}
	else /* job found */
	    start_time_job += proc_time[job];
    }

    sequence.value = compute_obj_function( );  
}



long int compute_obj_function( ) {

/* Computes the objective function value of sequence k for the SMWC from scratch */

    long int  i;
    long int  obj_f_value = 0;
    long int  start_time_job, end_time_job, job;

   /* first compute the completion times of all jobs */

   /* initialize for the first job */
    job = sequence.job_at_pos[0];
    start_time_job = release_date[job];
    end_time_job = start_time_job + proc_time[job];
    sequence.start_time_job[job] = start_time_job;
    sequence.completion_time_job[job] = end_time_job;
    sequence.completion_time_pos[0] = end_time_job;
    /* now the remaining jobs */
    for ( i = 1 ; i < number_jobs ; i++ ) {
	job = sequence.job_at_pos[i];
	start_time_job = MAX(release_date[job],sequence.completion_time_pos[i-1]);
	sequence.start_time_job[job] = start_time_job;
	end_time_job = start_time_job + proc_time[job];
	sequence.completion_time_job[job] = end_time_job;
	sequence.completion_time_pos[i] = end_time_job;
    }

    /* Now compute the objective function value */
    for ( i = 0 ; i < number_jobs ; i++ ) 
	obj_f_value += sequence.completion_time_job[i] * weight[i];

    return( obj_f_value );
}



void check_solution_consisteny_with_release_dates(long int k) {
    long int  i;

    for ( i = 0 ; i < number_jobs ; i++ ) {
	assert (sequence.start_time_job[sequence.job_at_pos[i]] >= release_date[i] );
    }
   
}



void generate_random_vector() {
    /* Generates a random vector: quick and dirty. */
    long int  i, help, job;
    long int  tot_assigned;
    float     rnd;

    for ( i=0 ; i <= number_jobs; i++) 
	random_vector[i] = i;
   
    tot_assigned = 0;
    for ( i=0 ; i < number_jobs; i++) 
    {
	/* find (randomly) an long int for a free unit */ 
	rnd  = ran01 ( &seed );
	job = (long int) (rnd  * (number_jobs - tot_assigned)); 
	assert( i + job < number_jobs );
	help = random_vector[i];
	random_vector[i] = random_vector[i+job];
	random_vector[i+job] = help;
	tot_assigned++;
    }
}


 
void local_search( ) {

    switch ( local_search_flag ) {
	case 0:
	    ; 
	    break;
	    
	case 1:
	    ls_transpose_first( );
	    break;
	case 2: 
	    ls_exchange_first( );
	    break;
	case 3: 
	    ls_insert_first( );
	    break;
	case 4: 
	    ls_exchange_first( );
	    ls_insert_first( );
	    break;
	case 5: 
	    ls_insert_first( );
	    ls_exchange_first( );
	    break;
	case 6: 
	    ls_transpose_first( );
	    ls_exchange_first( );
	    ls_insert_first( );
	    break;
	case 7: 
	    ls_transpose_first( );
	    ls_insert_first( );
	    ls_exchange_first( );
	    break;
    }
}



void ls_transpose_first( ) { 
/* this is a local search procedure based on the transpose
   neighbourhood. It uses a first-improvement pivoting rule and stops
   in the first local optimum encountered. */

    long int i, job;
    long int suc_job;
    long int improvement_flag, difference;
    long int obj_f_value;

    VERBOSE ( printf("Before local search with transpose-neigh. weighted Compl.time: %ld\n",sequence.value); )
    improvement_flag = TRUE;
    while (improvement_flag) {  
	improvement_flag = FALSE;
	for (i = 0; i < number_jobs - 1 ; i++) {
	    job = sequence.job_at_pos[i];
	    suc_job = sequence.job_at_pos[i+1];
	    sequence.job_at_pos[i]   = suc_job;
	    sequence.job_at_pos[i+1] = job;
	    sequence.pos_of_job[job] = i+1;
	    sequence.pos_of_job[suc_job] = i;
	    /* I know that recomputing from scratch is costly; but
               this is only an extremely basic code */
	    obj_f_value = compute_obj_function( ); 
	    difference = obj_f_value - sequence.value;
	    if (difference < 0) {
		improvement_flag = TRUE;
		sequence.value += difference;
	    }
	    else {
		sequence.job_at_pos[i] = job;
		sequence.job_at_pos[i+1] = suc_job;
		sequence.pos_of_job[job] = i;
		sequence.pos_of_job[suc_job] = i+1;
	    }
	}
    }
    VERBOSE ( printf("After local search with transpose-neigh. weighted Compl.time: %ld\n",sequence.value); );
}



void ls_exchange_first( ) {
/* this is a local search procedure based on the exchange
   neighbourhood. It uses a first-improvement pivoting rule and stops
   in the first local optimum encountered. */

    long int i, j, job_i, job_j;
    long int improvement_flag, difference;
    long int obj_f_value;

    VERBOSE ( printf("Before local search with exchange-neigh. weighted Compl.time: %ld\n",sequence.value); )
    improvement_flag = TRUE;
    while (improvement_flag) {  
	improvement_flag = FALSE;
	for (i = 0; i < number_jobs - 1 ; i++) {
	    for (j = i+1; j < number_jobs ; j++) {
		job_i = sequence.job_at_pos[i];
		job_j = sequence.job_at_pos[j];
		sequence.job_at_pos[i]   = job_j;
		sequence.job_at_pos[j] = job_i;
		sequence.pos_of_job[job_i] = j;
		sequence.pos_of_job[job_j] = i;
		obj_f_value = compute_obj_function( ); 
		difference = obj_f_value - sequence.value;
		if (difference < 0) {
		    improvement_flag = TRUE;
		    sequence.value += difference;
		}
		else {
		    sequence.job_at_pos[i] = job_i;
		    sequence.job_at_pos[j] = job_j;
		    sequence.pos_of_job[job_i] = i;
		    sequence.pos_of_job[job_j] = j;
		}
	    }
	}
    }
    VERBOSE ( printf("After local search with exchange-neigh. weighted Compl.time: %ld\n",sequence.value); );
}



void ls_insert_first( ) {
/* this is a local search procedure based on the insert
   neighbourhood. It uses a first-improvement pivoting rule and stops
   in the first local optimum encountered. */

    long int i, j, l, job_i;
    long int improvement_flag, difference = 0;
    long int obj_f_value;

    VERBOSE ( printf("Before local search with insert-neigh. weighted Compl.time: %ld\n",sequence.value); )
    improvement_flag = TRUE;
    while (improvement_flag) {  
	improvement_flag = FALSE;
	for (i = 0; i < number_jobs ; i++) {
	    for (j = 0; j < number_jobs ; j++) {
		job_i = sequence.job_at_pos[i];
		if ( i == j )
		    continue;
		else if ( i < j ) {
		    for ( l = i ; l < j ; l++ ) {
			sequence.job_at_pos[l] = sequence.job_at_pos[l+1];
			sequence.pos_of_job[sequence.job_at_pos[l]] = l;
		    }
		    sequence.job_at_pos[j] = job_i;
		    sequence.pos_of_job[job_i] = j;
		    obj_f_value = compute_obj_function( ); 
		    difference = obj_f_value - sequence.value;
		}
		else if ( i > j ) {
		    for ( l = i ; l > j ; l-- ) {
			sequence.job_at_pos[l] = sequence.job_at_pos[l-1];
			sequence.pos_of_job[sequence.job_at_pos[l]] = l;
		    }
		    sequence.job_at_pos[j] = job_i;
		    sequence.pos_of_job[job_i] = j;
		    obj_f_value = compute_obj_function( ); 
		    difference = obj_f_value - sequence.value;
		}
		if (difference < 0) {
		    improvement_flag = TRUE;
		    sequence.value += difference;
		}
		else {
		    /* undo change */
		    if ( i < j ) {
			for ( l = j ; l > i ; l-- ) {
			    sequence.job_at_pos[l] = sequence.job_at_pos[l-1];
			    sequence.pos_of_job[sequence.job_at_pos[l]] = l;
			}
			sequence.job_at_pos[i] = job_i;
			sequence.pos_of_job[job_i] = i;
		    }
		    else if ( i > j ) {
			for ( l = j ; l < i ; l++ ) {
			    sequence.job_at_pos[l] = sequence.job_at_pos[l+1];
			    sequence.pos_of_job[sequence.job_at_pos[l]] = l;
			}
			sequence.job_at_pos[i] = job_i;
			sequence.pos_of_job[job_i] = i;
		    }
		}
	    }
	}
    }
    VERBOSE ( printf("After local search with insert-neigh. weighted Compl.time: %ld\n",sequence.value); )
	}

